# ========================================================================================================
# Mark Neeser
# 2015.07.01
#
# A python interactive window for use in the Austrian in-kind
# fringe-correction Reflex workflow.  This interactive window
# shows the results of the fringe-correction pipeline, allowing
# the User to compare the input image with the fringe-corrected
# image, and the fringe map.   From this interactive Actor the
# User can modify pipeline parameters and re-initiate the
# processing.
#
# Pipeline Parameters for inclusion:
#
# 1. esotk_create_mask
#    --sigma_det      object detection threshold in units of standard deviation [3.0]
#    --growing_radius radius by which mask of detected object should be grown (<= 100) [5]
#
# 2. esotk_masterfringe_create
#    --rescale        measure fringe amplitude and rescale images before stacking? [yes]
#    --nmin_comb      minimum number of jittered images to combine [1]
#    --method         method for computing a robust mean: median|min_max_rejection|
#                     kappa_sigma_clip|mean_median_ratio_constraint|average|sum) [median]
#    --rejlow         the number of low-level frames to reject [3]
#    --rejhigh        the number of high-level frames to reject [3]
#    --kappa_low      lower side rejection kappa multiple [3.0]
#    --kappa_high     high side rejection kappa multiple [3.0]
#    --ks_iter        number of iterations of kappa sigma clipping for running computation of mean [5]
#    --mmr_threshold  threshold for mean/median ratio [1.001]
#
# 3. esotk_masterfringe_correct
#    --rescale        if "yes", then the fringe amplitude is measured in each input image and the fringe image
#                     is scaled accordingly before subtracting.  [yes]
#
#
# Relevant output images to plot (and the recipes that created them)
#
# 1. OBJ_MASK         (esotk_mask_create)      ==> mask of sources in image
# 2. MASTER_FRINGE    (esotk_masterfringe_create)    ==> pure fringe map (no sources)
# 3. FRINGE_CORRECTED (esotk_masterfringe_correct)   ==> fringe-corrected input image
#
#
# ========================================================================================================

from __future__ import absolute_import
import os
import sys
sys.path.append(os.path.dirname(__file__))
from esotk_fringe_gui import BaseDataPlotterManager, runInteractiveApp

try:
    # Import non-standard but needed modules. These must be as subset of what
    # the esotk_fringe_gui module tries to load.
    import reflex
    import pylab
    import pipeline_product
    import pipeline_display

    # This class deals with the specific details of data reading and final
    # plotting.
    class DataPlotterManager(BaseDataPlotterManager):

        # This function will read all the columns, images and whatever is needed
        # from the products. The variables , self.plot_x, self.plot_y, etc...
        # are used later in function plotProductsGraphics().
        def readFitsData(self, fitsFiles):
            # Control variable to check if the interesting files were found in
            # the input.
            self.science_frames = []
            self.science2_frames = []
            fringe_frame = None

            # The following will alias science1_frames to science_frames.
            self.science1_frames = self.science_frames

            # Read all the products
            for frame in fitsFiles:
                if frame == '':
                    continue
                category = frame.category
                if category == 'RAW':
                    self.science1_frames.append(frame)
                elif category == 'MASTER_FRINGE':
                    fringe_frame = frame
                elif category == 'FRINGE_CORRECTED':
                    self.science2_frames.append(frame)

            self.science1_found = len(self.science1_frames) > 0
            self.fringe_found = fringe_frame is not None
            self.science2_found = len(self.science2_frames) > 0

            if self.science1_found:
                frame = self.science1_frames[0]
                self.qc_index = 1
                self.science1 = pipeline_product.PipelineProduct(frame)
                self.setupExtensionValues(self.science1)

            if self.fringe_found:
                self.fringe = pipeline_product.PipelineProduct(fringe_frame)

            if self.science2_found:
                frame = self.science2_frames[0]
                self.science2 = pipeline_product.PipelineProduct(frame)

        def loadProducts(self, index):
            """
            Loading both uncorrected and corrected science frames.
            Also remember the suffix number of the QC parameters to display for
            the uncorrected science frames.
            """
            super(DataPlotterManager, self).loadProducts(index)
            self.qc_index = index + 1
            self.science1 = self.science
            if index >= 0 and index < len(self.science2_frames):
                frame = self.science2_frames[index]
                self.science2 = pipeline_product.PipelineProduct(frame)

        # This function creates all the subplots. It is responsible for the
        # plotting layouts.
        # There can different layouts, depending on the availability of data
        # Note that subplot(I,J,K) means the Kth plot in a IxJ grid
        # Note also that the last one is actually a box with text, no graphs.
        def addSubplots(self, figure):
            if self.science1_found == True and self.science2_found == True and self.fringe_found == True:
                self.subplot_image1 = figure.add_subplot(1, 3, 1)
                self.subplot_image2 = figure.add_subplot(1, 3, 2)
                self.subplot_image3 = figure.add_subplot(1, 3, 3)
            else:
                self.subtext_nodata = figure.add_subplot(1, 1, 1)
            self.bkgLevelLabel1 = figure.text(0.1, 0.91,
                                              "Input image QC parameters:")
            self.bkgLevelLabel1 = figure.text(0.1, 0.88,
                                              "Background level = Not found")
            self.fringeAmpLabel1 = figure.text(0.1, 0.85,
                                               "Fringe amplitude = Not found")
            self.bkgLevelLabel2 = figure.text(0.725, 0.91,
                                              "Corrected image QC parameters:")
            self.bkgLevelLabel2 = figure.text(0.725, 0.88,
                                              "Background level = Not found")
            self.fringeAmpLabel2 = figure.text(0.725, 0.85,
                                               "Fringe amplitude = Not found")
            self.setupExtensionLabel(figure)


        # This is the function that makes the plots.
        # Add new plots or delete them using the given scheme.
        # The data has been already stored in self.plot_x, self.plot_xdif, etc ...
        # It is mandatory to add a tooltip variable to each subplot.
        # One might be tempted to merge addSubplots() and plotProductsGraphics().
        # There is a reason not to do it: addSubplots() is called only once at
        # startup, while plotProductsGraphics() is called always there is a
        # resize.
        def plotProductsGraphics(self):

            if self.science1_found == True and self.science2_found == True and self.fringe_found == True:

                # Display input image and masked image
                
                title_image1 = 'Input Image ext['+str(self.ext_id)+']'
                tooltip_image1 = """This is the input image with fringes"""
                
                title_image2 = 'Fringe Map ext['+str(self.ext_id)+']'
                tooltip_image2 = """This is the computed fringe map"""
                
                title_image3 = 'Corrected Image ext['+str(self.ext_id)+']'
                tooltip_image3 = """This is the input image corrected with a scaled fringe map"""
                
                ext_id=self.ext_id
                self.science1.readImage(fits_extension=ext_id)
                self.fringe.readImage(fits_extension=ext_id)
                self.science2.readImage(fits_extension=ext_id)
                imgdisp = pipeline_display.ImageDisplay()
                imgdisp.setAspect('equal')  # this sets images to an equal aspect ratio

                # using a MAD to find reasonable display limits for the input images:
                image1   = self.science1.image[ext_id]
                new_med1, new_mad1 = self.computeRobustMedMad(image1)

                image2   = self.fringe.image[ext_id]
                new_med2, new_mad2 = self.computeRobustMedMad(image2)

                image3   = self.science2.image[ext_id]
                new_med3, new_mad3 = self.computeRobustMedMad(image3)

                imgdisp.z_lim = new_med1 - 1.48*new_mad1, new_med1 + 3*1.48*new_mad1
                imgdisp.setCmap('hot')
                imgdisp.setLabels('X (pixels)', 'Y (pixels)')
                imgdisp.display(self.subplot_image1, title_image1, tooltip_image1, self.science1.image)

                # Set limits to mask image
                imgdisp.z_lim = new_med2 - 1.48*new_mad2, new_med2 + 3*1.48*new_mad2
                pylab.setp(self.subplot_image2.get_yticklabels(), visible = False)
                imgdisp.setCmap('hot')
                imgdisp.setLabels('X (pixels)', ' ')
                imgdisp.display(self.subplot_image2, title_image2, tooltip_image2, self.fringe.image)

                # Set limits to mask image
                imgdisp.z_lim = new_med3 - 1.48*new_mad3, new_med3 + 3*1.48*new_mad3
                pylab.setp(self.subplot_image3.get_yticklabels(), visible = False)
                imgdisp.setCmap('hot')
                imgdisp.setLabels('X (pixels)', ' ')
                imgdisp.display(self.subplot_image3, title_image3, tooltip_image3, self.science2.image)
                
                self.subplot_image1.set_title(title_image1, fontsize=12, fontweight='semibold', rotation='vertical',x=1.05,y=0.6)
                self.subplot_image2.set_title(title_image2, fontsize=12, fontweight='semibold', rotation='vertical',x=1.05,y=0.6)
                self.subplot_image3.set_title(title_image3, fontsize=12, fontweight='semibold', rotation='vertical',x=1.05,y=0.6)

                self.updateQCparamLabels()
                self.updateExtensionLabel()

            else:
                # Data not found info
                self.subtext_nodata.set_axis_off()
                self.text_nodata = """Data not found. Input files should contain these types:
RAW
MASTER_FRINGE
FRINGE_CORRECTED
"""
                self.subtext_nodata.text(0.1, 0.6, self.text_nodata,
                                         color = '#11557c', fontsize = 18,
                                         ha = 'left', va = 'center', alpha = 1.0)
                self.subtext_nodata.tooltip = "Data not found in the products"

        def updateQCparamLabels(self):
            """
            Extracts the QC parameters for the input and corrected raw frames
            and update the plot labels. Note: QC parameters for the raw
            uncorrected input frames are loaded from the master fringe.
            """

            def _getParamFromFringe(keyword):
                # Fits try load from a numbered keyword. If that cannot be found
                # then assume there was only one raw input file and load from an
                # unnumbered keyword instead.
                full_keyword = "{0}{1}".format(keyword, self.qc_index)
                try:
                    return self.fringe.readKeyword(full_keyword, self.ext_id)
                except KeyError:
                    try:
                        return self.fringe.readKeyword(keyword, self.ext_id)
                    except KeyError:
                        return "Not found"

            value = _getParamFromFringe("ESO QC BKGLEVEL")
            text = "Background level = {0:g}".format(value)
            self.bkgLevelLabel1.set_text(text)
            value = _getParamFromFringe("ESO QC FRINGEAMP")
            text = "Fringe amplitude = {0:g}".format(value)
            self.fringeAmpLabel1.set_text(text)

            try:
                value = self.science2.readKeyword("ESO QC BKGLEVEL",
                                                  self.ext_id)
                text = "Background level = {0:g}".format(value)
                self.bkgLevelLabel2.set_text(text)
                value = self.science2.readKeyword("ESO QC FRINGEAMP",
                                                  self.ext_id)
                text = "Fringe amplitude = {0:g}".format(value)
                self.fringeAmpLabel2.set_text(text)
            except KeyError:
                self.bkgLevelLabel2.set_text("None found.")
                self.fringeAmpLabel2.set_text("Rescaling not applied.")

        # This function specifies which are the parameters that should be presented
        # in the window to be edited.
        # Note that the parameter has to be also in the in_sop port (otherwise it
        # won't appear in the window)
        # The descriptions are used to show a tooltip. They should match one to one
        # with the parameter list
        # Note also that parameters have to be prefixed by the 'recipe name:'
        def setInteractiveParameters(self):
            paramList = list()

            rec_id = "esotk_masterfringe_correct"
            group_id = "extra"
            
            pl = self.interactive_app.inputs.in_sop
            for p in pl:
                
                p_name=p.displayName
                
                p_desc=p.description

                if p.valenum :
                    p_desc += ' <'
                    for value in p.valenum :  
                        p_desc += value + ' |'
                    p_desc = p_desc[:-1] + "> [" +p.value + "]"
                
                paramList.append(reflex.RecipeParameter(
                     rec_id, p_name, group=group_id, description=p_desc))
                

            return paramList


        def setWindowHelp(self):
            help_text = """
This is an interactive window which help to assess the quality of the execution of a recipe.
"""
            return help_text

        def setWindowTitle(self):
            title = 'Pipeline Interactive Window'
            return title

except ImportError:
    DataPlotterManager = None


if __name__ == '__main__':
    runInteractiveApp(DataPlotterManager)
